@@ -98,7 +98,7 @@ class ClientSession(object):
     :param extra_credential: extra credentials. as list of ``(key, value)``
                              tuples.
     :param client_tags: Client tags as list of strings.
-    :param role: role for the current session. Some connectors do not
+    :param roles: roles for the current session. Some connectors do not
                  support role management. See connector documentation for more details.
     """
 
@@ -113,7 +113,7 @@ class ClientSession(object):
         transaction_id: str = None,
         extra_credential: List[Tuple[str, str]] = None,
         client_tags: List[str] = None,
-        role: str = None,
+        roles: Dict[str, str] = None,
     ):
         self._user = user
         self._catalog = catalog
@@ -124,7 +124,8 @@ class ClientSession(object):
         self._transaction_id = transaction_id
         self._extra_credential = extra_credential
         self._client_tags = client_tags
-        self._role = role
+        self._roles = roles or {}
+        self._prepared_statements: Dict[str, str] = {}
         self._object_lock = threading.Lock()
 
     @property
@@ -187,6 +188,25 @@ class ClientSession(object):
     def client_tags(self):
         return self._client_tags
 
+    @property
+    def roles(self):
+        with self._object_lock:
+            return self._roles
+
+    @roles.setter
+    def roles(self, roles):
+        with self._object_lock:
+            self._roles = roles
+
+    @property
+    def prepared_statements(self):
+        return self._prepared_statements
+
+    @prepared_statements.setter
+    def prepared_statements(self, prepared_statements):
+        with self._object_lock:
+            self._prepared_statements = prepared_statements
+
     def __getstate__(self):
         state = self.__dict__.copy()
         del state["_object_lock"]
@@ -196,16 +216,6 @@ class ClientSession(object):
         self.__dict__.update(state)
         self._object_lock = threading.Lock()
 
-    @property
-    def role(self):
-        with self._object_lock:
-            return self._role
-
-    @role.setter
-    def role(self, role):
-        with self._object_lock:
-            self._role = role
-
 
 def get_header_values(headers, header):
     return [val.strip() for val in headers[header].split(",")]
@@ -214,7 +224,23 @@ def get_header_values(headers, header):
 def get_session_property_values(headers, header):
     kvs = get_header_values(headers, header)
     return [
-        (k.strip(), urllib.parse.unquote(v.strip()))
+        (k.strip(), urllib.parse.unquote_plus(v.strip()))
+        for k, v in (kv.split("=", 1) for kv in kvs)
+    ]
+
+
+def get_prepared_statement_values(headers, header):
+    kvs = get_header_values(headers, header)
+    return [
+        (k.strip(), urllib.parse.unquote_plus(v.strip()))
+        for k, v in (kv.split("=", 1) for kv in kvs)
+    ]
+
+
+def get_roles_values(headers, header):
+    kvs = get_header_values(headers, header)
+    return [
+        (k.strip(), urllib.parse.unquote_plus(v.strip()))
         for k, v in (kv.split("=", 1) for kv in kvs)
     ]
 
@@ -382,7 +408,12 @@ class TrinoRequest(object):
         headers[constants.HEADER_SCHEMA] = self._client_session.schema
         headers[constants.HEADER_SOURCE] = self._client_session.source
         headers[constants.HEADER_USER] = self._client_session.user
-        headers[constants.HEADER_ROLE] = self._client_session.role
+        if len(self._client_session.roles.values()):
+            headers[constants.HEADER_ROLE] = ",".join(
+                # ``name`` must not contain ``=``
+                "{}={}".format(catalog, urllib.parse.quote(str(role)))
+                for catalog, role in self._client_session.roles.items()
+            )
         if self._client_session.client_tags is not None and len(self._client_session.client_tags) > 0:
             headers[constants.HEADER_CLIENT_TAGS] = ",".join(self._client_session.client_tags)
 
@@ -392,6 +423,13 @@ class TrinoRequest(object):
             for name, value in self._client_session.properties.items()
         )
 
+        if len(self._client_session.prepared_statements) != 0:
+            # ``name`` must not contain ``=``
+            headers[constants.HEADER_PREPARED_STATEMENT] = ",".join(
+                "{}={}".format(name, urllib.parse.quote_plus(statement))
+                for name, statement in self._client_session.prepared_statements.items()
+            )
+
         # merge custom http headers
         for key in self._client_session.headers:
             if key in headers.keys():
@@ -554,7 +592,22 @@ class TrinoRequest(object):
             self._client_session.schema = http_response.headers[constants.HEADER_SET_SCHEMA]
 
         if constants.HEADER_SET_ROLE in http_response.headers:
-            self._client_session.role = http_response.headers[constants.HEADER_SET_ROLE]
+            for key, value in get_roles_values(
+                    http_response.headers, constants.HEADER_SET_ROLE
+            ):
+                self._client_session.roles[key] = value
+
+        if constants.HEADER_ADDED_PREPARE in http_response.headers:
+            for name, statement in get_prepared_statement_values(
+                http_response.headers, constants.HEADER_ADDED_PREPARE
+            ):
+                self._client_session.prepared_statements[name] = statement
+
+        if constants.HEADER_DEALLOCATED_PREPARE in http_response.headers:
+            for name in get_header_values(
+                http_response.headers, constants.HEADER_DEALLOCATED_PREPARE
+            ):
+                self._client_session.prepared_statements.pop(name)
 
         self._next_uri = response.get("nextUri")
 
@@ -592,33 +645,35 @@ class TrinoResult(object):
     https://docs.python.org/3/library/stdtypes.html#generator-types
     """
 
-    def __init__(self, query, rows=None):
+    def __init__(self, query, rows: List[Any]):
         self._query = query
-        self._rows = rows or []
+        # Initial rows from the first POST request
+        self._rows = rows
         self._rownumber = 0
 
+    @property
+    def rows(self):
+        return self._rows
+
+    @rows.setter
+    def rows(self, rows):
+        self._rows = rows
+
     @property
     def rownumber(self) -> int:
         return self._rownumber
 
     def __iter__(self):
-        # Initial fetch from the first POST request
-        for row in self._rows:
-            self._rownumber += 1
-            yield row
-        self._rows = None
-
-        # Subsequent fetches from GET requests until next_uri is empty.
-        while not self._query.finished:
-            rows = self._query.fetch()
-            for row in rows:
+        # A query only transitions to a FINISHED state when the results are fully consumed:
+        # The reception of the data is acknowledged by calling the next_uri before exposing the data through dbapi.
+        while not self._query.finished or self._rows is not None:
+            next_rows = self._query.fetch() if not self._query.finished else None
+            for row in self._rows:
                 self._rownumber += 1
                 logger.debug("row %s", row)
                 yield row
 
-    @property
-    def response_headers(self):
-        return self._query.response_headers
+            self._rows = next_rows
 
 
 class TrinoQuery(object):
@@ -641,8 +696,7 @@ class TrinoQuery(object):
         self._request = request
         self._update_type = None
         self._sql = sql
-        self._result = TrinoResult(self)
-        self._response_headers = None
+        self._result: Optional[TrinoResult] = None
         self._experimental_python_types = experimental_python_types
         self._row_mapper: Optional[RowMapper] = None
 
@@ -652,7 +706,7 @@ class TrinoQuery(object):
             while not self._columns and not self.finished and not self.cancelled:
                 # Columns are not returned immediately after query is submitted.
                 # Continue fetching data until columns information is available and push fetched rows into buffer.
-                self._result._rows += self.fetch()
+                self._result.rows += self.fetch()
         return self._columns
 
     @property
@@ -697,8 +751,11 @@ class TrinoQuery(object):
             self._finished = True
 
         rows = self._row_mapper.map(status.rows) if self._row_mapper else status.rows
-
         self._result = TrinoResult(self, rows)
+
+        # Execute should block until at least one row is received or query is finished or cancelled
+        while not self.finished and not self.cancelled and len(self._result.rows) == 0:
+            self._result.rows += self.fetch()
         return self._result
 
     def _update_state(self, status):
@@ -716,7 +773,6 @@ class TrinoQuery(object):
         status = self._request.process(response)
         self._update_state(status)
         logger.debug(status)
-        self._response_headers = response.headers
         if status.next_uri is None:
             self._finished = True
 
@@ -754,10 +810,6 @@ class TrinoQuery(object):
     def cancelled(self) -> bool:
         return self._cancelled
 
-    @property
-    def response_headers(self):
-        return self._response_headers
-
 
 def _retry_with(handle_retry, handled_exceptions, conditions, max_attempts):
     def wrapper(func):
@@ -921,8 +973,7 @@ class RowMapper:
     """
     Maps a row of data given a list of mapping functions
     """
-
-    def __init__(self, columns=[]):
+    def __init__(self, columns):
         self.columns = columns
 
     def map(self, rows):
